import os
from functools import partial
from io import BytesIO

from PIL import Image as pilim
from django.db.models.fields.files import FieldFile
from reportlab.lib import colors
from reportlab.lib.enums import TA_LEFT
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.utils import ImageReader
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.platypus import BaseDocTemplate, Image
from reportlab.platypus import Frame, PageTemplate
from reportlab.platypus import Paragraph, Table, TableStyle

styles = getSampleStyleSheet()
styleN = styles['Normal']
styleH = styles['Heading1']

DESIRED_PHOTO_WIDTH = A4[0] - 100
DESIRED_PHOTO_HEIGHT = A4[1] / 2 - 170

BODY_FONT_SIZE = 8
BASIC_MARGIN = 30
PAGE_WIDTH = A4[0]
PAGE_HEIGHT = A4[1]
FRAME_TOP = 120
__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


class MCPDFBuilder:

    def __init__(self, basic_margin=BASIC_MARGIN, top_margin=BASIC_MARGIN, right_margin=BASIC_MARGIN,
                 bottom_margin=BASIC_MARGIN, left_margin=BASIC_MARGIN, page_width=PAGE_WIDTH, page_height=PAGE_HEIGHT,
                 logo_path=None, watermark=None, company_info=None, pagesize=A4):
        self.pdf_buffer = BytesIO()
        self.pdf = BaseDocTemplate(self.pdf_buffer, pagesize=pagesize)
        frame = Frame(
            x1=left_margin,
            y1=bottom_margin,
            width=page_width - left_margin - right_margin,
            height=page_height - top_margin - FRAME_TOP,
            showBoundary=1
        )
        header_and_footer_content = {
            'watermark': watermark,
            'logo_path': logo_path,
            'body_style': get_body_style(),
            'company_info': company_info,
            'body_font_size': BODY_FONT_SIZE,
            'basic_margin': basic_margin,
            'top_margin': top_margin,
            'right_margin': right_margin,
            'bottom_margin': bottom_margin,
            'left_margin': left_margin,
            'page_width': page_width,
            'page_height': page_height,
            'line_y': page_height,
        }
        template = PageTemplate(
            id='all_pages', frames=frame,
            onPage=partial(header_and_footer, content=header_and_footer_content)
        )
        self.pdf.addPageTemplates([template])

    def add_to_story(self, story, obj):
        if isinstance(obj, list):
            for x in obj:
                story.append(x)
        else:
            story.append(obj)
        return story

    def build(self, story):
        self.pdf.build(story)
        pdf_value = self.pdf_buffer.getvalue()
        self.pdf_buffer.close()
        return pdf_value


class Row:
    def __init__(self, columns, grid=None, background_color=None, text_color=None, full_width=False, align=TA_LEFT):
        self.columns = columns
        self.grid = grid
        self.background_color = background_color
        self.text_color = text_color
        self.align = align
        self.full_width = full_width

    def get_number_of_columns(self):
        return len(self.columns)


class Column:

    def __init__(self, content, grid=None, background_color=None, text_color=None, image_width=50, image_height=50,
                 align=TA_LEFT, col_width=None):
        """

        :param content: This variable's data will be passed into the Reportlab's Paragraph Class, or if its BytesIO
                        object will be passed in a Reportlab's Image Class object and appended in the table.
        :param grid:
        :param background_color:
        :param text_color:
        :param image_width:
        :param image_height:
        :param align: Align options (from ReportLab) -> TA_LEFT, TA_CENTER, TA_RIGHT, TA_JUSTIFY
        """
        self.content = content
        self.grid = grid
        self.background_color = background_color
        self.text_color = text_color
        self.image_width = image_width
        self.image_height = image_height
        self.align = align


class Grid:
    def __init__(self, line_width=0.5, line_color=colors.gray):
        self.line_width = line_width
        self.line_color = line_color


def get_img(img_path):
    # buff1 = BytesIO()
    # img = pilim.open(img_path)
    # print('img_path:', img_path, img.size[0], img.size[1])
    # door_out_img = Image(img_path, width=2 * inch, height=2 * inch)

    return '''<para autoLeading="off" fontSize=12>This &lt;img/&gt; <img
src="{0}" valign="top" width="{1}" height="{2}"/> </para>'''.format(img_path, 100, 100)


def pfd_table_builder(data, col_widths=None):
    """
    :param data:
    :return: A reportlab table instance
    """
    columns_counter_final = 0
    table_data = []
    for row_ctr, row_obj in enumerate(data):
        if isinstance(row_obj, Row):
            column_final = []
            columns_counter = 0
            for col_ctr, col in enumerate(row_obj.columns):
                columns_counter += 1
                style = get_body_style()
                # print('{}, {}'.format(row_ctr, col_ctr))
                # Loop through all the cells and apply the styles on cell level if declared
                # specifically for the particular cell or else apply the style of the row on the cell (if any).
                # Align
                if col.align:
                    style.alignment = col.align
                elif row_obj.align:
                    style.alignment = row_obj.align
                # Text Color
                if col.text_color:
                    style.textColor = col.text_color
                elif row_obj.text_color:
                    style.textColor = row_obj.text_color

                # print("type:", type(col.content))
                if isinstance(col.content, BytesIO) or isinstance(col.content, FieldFile):
                    _col = Image(col.content, width=col.image_width, height=col.image_height)
                elif isinstance(col.content, str):
                    _col = Paragraph('{}'.format(col.content, 'N/A'), style)
                else:
                    # print("SOMETHING ELSE")
                    _col = col.content

                column_final.append(_col)

            table_data.append(column_final)
        else:
            table_data.append([row_obj])

        if columns_counter > columns_counter_final:
            columns_counter_final = columns_counter

    # ------------------ STYLE
    table_styles = TableStyle()
    for row_ctr, row_obj in enumerate(data):
        if isinstance(row_obj, Row):
            # Loop through all the cells and apply the styles on cell level if declared
            # specifically for the particular cell or else apply the style of the row on the cell
            # (so every cell on this row will have this style, if any).
            for col_ctr, col_obj in enumerate(row_obj.columns):
                apply_style = False
                # Background Color
                background_color = None
                if col_obj.background_color:
                    apply_style = True
                    background_color = col_obj.background_color
                elif row_obj.background_color:
                    apply_style = True
                    background_color = row_obj.background_color
                if apply_style:
                    table_styles.add('BACKGROUND', (col_ctr, row_ctr), (col_ctr, row_ctr), background_color)
                # Grid
                apply_style = False
                grid_line_color = None
                grid_line_width = None
                if col_obj.grid:
                    apply_style = True
                    grid_line_color = col_obj.grid.line_color
                    grid_line_width = col_obj.grid.line_width
                elif row_obj.grid:
                    apply_style = True
                    grid_line_color = row_obj.grid.line_color
                    grid_line_width = row_obj.grid.line_width
                if apply_style:
                    table_styles.add('GRID', (col_ctr, row_ctr), (col_ctr, row_ctr), grid_line_width, grid_line_color)
                # SPAN
                if row_obj.full_width:
                    table_styles.add('SPAN', (col_ctr, row_ctr), (-1, row_ctr))

    print('columns_counter_final:', columns_counter_final)
    t = Table(table_data, colWidths=col_widths)
    t.setStyle(table_styles)
    return t


def resize_pilim_and_keep_aspect_ratio(im, basewidth):
    wpercent = (basewidth / float(im.size[0]))
    hsize = int((float(im.size[1]) * float(wpercent)))
    im = im.resize((basewidth, hsize), pilim.ANTIALIAS)
    return im


def header_and_footer(canvas, pdf, content):
    # ----------------------- HEADER ----------------------- #
    watermark = content.get('watermark')
    logo_path = content.get('logo_path')
    body_style = content.get('body_style')
    company_info = content.get('company_info')
    body_font_size = content.get('body_font_size')
    basic_margin = content.get('basic_margin')
    top_margin = content.get('top_margin')
    right_margin = content.get('right_margin')
    bottom_margin = content.get('bottom_margin')
    left_margin = content.get('left_margin')
    page_width = content.get('page_width')
    page_height = content.get('page_height')
    line_y = content.get('line_y')

    if company_info and company_info.get('pdf_tab_title'):
        canvas.setTitle(company_info.get('pdf_tab_title'))

    # LOGO
    if company_info and company_info.get('logo'):
        im = pilim.open(company_info.get('logo'))
        img_width, img_height = im.size
        print_width, print_height = resize_img_and_keep_aspect_ratio(
            img_width, img_height, img_desired_height=FRAME_TOP - top_margin
        )
        im = resize_pilim_and_keep_aspect_ratio(im, print_width)
        pil_img = ImageReader(im)
        canvas.drawImage(
            pil_img,
            page_width - right_margin - im.size[0],
            page_height - im.size[1] - 10,
            width=im.size[0],
            height=im.size[1],
            mask='auto'
        )

    # canvas.line(left_margin, page_height - 10, page_width + right_margin, page_height - 300)

    # COMPANY'S INFO
    if company_info and company_info.get('pdf_top_left_corner_info'):
        pdf_top_left_corner_info = company_info.get('pdf_top_left_corner_info')
        canvas.setFont('Helvetica-Bold', 10)
        top_y = page_height - bottom_margin
        canvas.drawString(left_margin, top_y, company_info.get('pdf_tab_title'))
        canvas.setFont('Helvetica', 9)
        i = 0
        for key, value in pdf_top_left_corner_info.items():
            if value:
                if key == 'iban':
                    value = "IBAN: {}".format(value)
                if i == 0:
                    canvas.drawString(left_margin, top_y - 20, value.strip())
                else:
                    canvas.drawString(left_margin, top_y - 20 - i * 12, value.strip())
                i += 1

    canvas.line(left_margin, page_height - FRAME_TOP + 10, page_width - right_margin, page_height - FRAME_TOP + 10)

    # ----------------------- FOOTER ----------------------- #
    canvas.setFont('Helvetica-Bold', 8)
    # page number
    page_number_text = "%d" % pdf.page
    canvas.drawCentredString(page_width / 2, 8, '-' + page_number_text + '-')
    # watermark
    if watermark:
        canvas.drawCentredString(page_width - 100, 8, watermark)

    canvas.line(left_margin, 20, page_width - right_margin, 20)


def get_body_style():
    sample_style_sheet = getSampleStyleSheet()
    body_style = sample_style_sheet['BodyText']
    body_style.fontSize = BODY_FONT_SIZE
    pdfmetrics.registerFont(TTFont('NotoSans', os.path.join(__location__, 'fonts/NotoSans-Regular.ttf')))
    pdfmetrics.registerFont(TTFont('NotoSansBold', os.path.join(__location__, 'fonts/NotoSans-Bold.ttf')))
    body_style.fontName = 'NotoSans'
    return body_style


def find_max_num_of_columns(data):
    max_num_of_cols = 0
    for row_counter, row in enumerate(data):
        num_of_cols = len(row.columns) if isinstance(row, Row) else 1
        if num_of_cols > max_num_of_cols:
            max_num_of_cols = num_of_cols
    return max_num_of_cols


def append_colon(param):
    return '{}:'.format(param)


def resize_img_and_keep_aspect_ratio(img_width, img_height, img_desired_width=None, img_desired_height=None):
    # print("img_width:", img_width)
    # print("img_height:", img_height)
    # print("img_desired_width:", img_desired_width)
    # print("img_desired_height:", img_desired_height)
    # Determine the dimensions of the image in the overview
    print_width = img_width
    print_height = img_height
    image_aspect = float(img_width) / float(img_height)
    # if photo exceeds desired width, resize to desired width
    if img_desired_width is not None:
        if img_width > img_desired_width:
            print_width = img_desired_width
            print_height = print_width / image_aspect
    # if photo exceeds desired height, resize to desire height
    if img_desired_height is not None:
        if print_height > img_desired_height:
            print_height = img_desired_height
            print_width = image_aspect * print_height

    # print("print_width:", print_width)
    # print("print_height:", print_height)
    # print("\n")
    return round(print_width), round(print_height)
